@@ -21,8 +21,16 @@
 #include <mousepad/mousepad-dialogs.h>
 #include <mousepad/mousepad-history.h>
 
+#ifdef HAVE_ERRNO_H
+#include <errno.h>
+#endif
 
 
+
+#ifndef HAVE_REALPATH
+#define realpath(path, resolved_path) NULL
+#endif
+
 enum
 {
   ENCODING_CHANGED,
@@ -67,8 +75,10 @@ struct _MousepadFile
 
   /* file monitoring */
   GFileMonitor *monitor;
+  GFile        *monitor_location;
   gchar        *etag;
-  gboolean      readonly;
+  gboolean      readonly, symlink;
+  guint         deleted_id;
 
   /* encoding of the file */
   MousepadEncoding encoding;
@@ -131,7 +141,10 @@ mousepad_file_init (MousepadFile *file)
   file->location = NULL;
   file->temporary = FALSE;
   file->monitor = NULL;
+  file->monitor_location = NULL;
   file->readonly = FALSE;
+  file->symlink = FALSE;
+  file->deleted_id = 0;
   file->encoding = mousepad_encoding_get_default ();
 #ifdef G_OS_WIN32
   file->line_ending = MOUSEPAD_EOL_DOS;
@@ -165,6 +178,9 @@ mousepad_file_finalize (GObject *object)
   if (file->monitor != NULL)
     g_object_unref (file->monitor);
 
+  if (file->monitor_location != NULL)
+    g_object_unref (file->monitor_location);
+
   if (file->autosave_location != NULL)
     g_object_unref (file->autosave_location);
 
@@ -188,6 +204,18 @@ mousepad_file_new (GtkTextBuffer *buffer)
   return file;
 }
 
+static gboolean
+mousepad_file_monitor_deleted (gpointer data)
+{
+  MousepadFile *file = data;
+
+  if (! mousepad_util_query_exists (file->monitor_location, FALSE))
+    gtk_text_buffer_set_modified (file->buffer, TRUE);
+
+  file->deleted_id = 0;
+
+  return FALSE;
+}
 
 
 static void
@@ -197,6 +225,8 @@ mousepad_file_monitor_changed (GFileMonitor      *monitor,
                                GFileMonitorEvent  event_type,
                                MousepadFile      *file)
 {
+  static gboolean deleted_pending = FALSE;
+
   GFileInfo *fileinfo;
 
   /* update readonly status */
@@ -213,14 +243,51 @@ mousepad_file_monitor_changed (GFileMonitor      *monitor,
            || event_type == G_FILE_MONITOR_EVENT_CREATED
            || event_type == G_FILE_MONITOR_EVENT_MOVED_IN
            || (event_type == G_FILE_MONITOR_EVENT_RENAMED
-               && g_file_equal (file->location, other_location)))
-    g_signal_emit (file, file_signals[EXTERNALLY_MODIFIED], 0);
-  /* the file has been deleted */
+               && g_file_equal (file->monitor_location, other_location)))
+    {
+      if (event_type != G_FILE_MONITOR_EVENT_CHANGED && file->deleted_id != 0)
+        {
+          g_source_remove (file->deleted_id);
+          file->deleted_id = 0;
+          deleted_pending = TRUE;
+
+          return;
+        }
+
+      g_signal_emit (file, file_signals[EXTERNALLY_MODIFIED], 0);
+
+      /* update monitor location in case of a symlink (exit this handler first) */
+      if (event_type != G_FILE_MONITOR_EVENT_CHANGED && (
+            file->symlink || (file->symlink = mousepad_util_is_symlink (file->location))
+          ))
+        g_idle_add (mousepad_file_set_monitor, mousepad_util_source_autoremove (file));
+
+      /* a "changed" event may or may not be issued following a "deleted"-"created" pair:
+       * if it is issued, we no longer wait for "done-hint" to issue it manually */
+      if (event_type == G_FILE_MONITOR_EVENT_CHANGED && deleted_pending)
+        deleted_pending = FALSE;
+    }
+  /*
+   * The file has been deleted: this can be permanently, or as a step in a modification
+   * process generating a "deleted" and then a "created" event. Unfortunately, there is no
+   * systematic way to differentiate between the two cases, so let's just give the "created"
+   * event a reasonable chance to reach us before deciding what action to take.
+   */
   else if (event_type == G_FILE_MONITOR_EVENT_DELETED
            || event_type == G_FILE_MONITOR_EVENT_MOVED_OUT
            || (event_type == G_FILE_MONITOR_EVENT_RENAMED
-               && g_file_equal (file->location, location)))
-    gtk_text_buffer_set_modified (file->buffer, TRUE);
+               && g_file_equal (file->monitor_location, location)))
+    {
+      if (file->deleted_id != 0)
+        g_source_remove (file->deleted_id);
+
+      file->deleted_id = g_timeout_add (MOUSEPAD_SETTING_GET_UINT (MONITOR_DISABLING_TIMER),
+                                        mousepad_file_monitor_deleted,
+                                        mousepad_util_source_autoremove (file));
+    }
+  else if (event_type == G_FILE_MONITOR_EVENT_CHANGES_DONE_HINT && deleted_pending)
+    mousepad_file_monitor_changed (monitor, location, NULL,
+                                   G_FILE_MONITOR_EVENT_CHANGED, file);
 }
 
 
@@ -229,9 +296,8 @@ static gboolean
 mousepad_file_set_monitor (gpointer data)
 {
   MousepadFile *file = data;
-  GFile        *location;
   GError       *error = NULL;
-  gchar        *path;
+  gchar        *path, *dir, *str;
 
   if (file->monitor != NULL)
     {
@@ -239,33 +305,76 @@ mousepad_file_set_monitor (gpointer data)
       file->monitor = NULL;
     }
 
+  if (file->monitor_location != NULL)
+    {
+      g_object_unref (file->monitor_location);
+      file->monitor_location = NULL;
+    }
+
   if (file->location != NULL && MOUSEPAD_SETTING_GET_BOOLEAN (MONITOR_CHANGES))
     {
       /* monitor the final target in case of a symlink: see
        * https://gitlab.gnome.org/GNOME/glib/-/issues/2421 */
-      if (g_file_query_file_type (file->location, G_FILE_QUERY_INFO_NOFOLLOW_SYMLINKS, NULL)
-            == G_FILE_TYPE_SYMBOLIC_LINK
-          && (path = realpath (mousepad_util_get_path (file->location), NULL)) != NULL)
+      if ((file->symlink = mousepad_util_is_symlink (file->location)))
         {
-          location = g_file_new_for_path (path);
-          g_free (path);
+          /* try to get the final target */
+          path = realpath (mousepad_util_get_path (file->location), NULL);
+
+          /* this is a broken link: we have to use readlink() to get the final target */
+          if (path == NULL && g_file_error_from_errno (errno) == G_FILE_ERROR_NOENT)
+            {
+              path = g_file_get_path (file->location);
+              dir = g_path_get_dirname (path);
+              while ((str = g_file_read_link (path, &error)) != NULL)
+                {
+                  g_free (path);
+                  if (g_str_has_prefix (str, "/"))
+                    path = str;
+                  else
+                    {
+                      path = g_strconcat (dir, "/", str, NULL);
+                      g_free (str);
+                    }
+                }
+
+              /* readlink() encountered a real error */
+              if (! g_error_matches (error, G_FILE_ERROR, G_FILE_ERROR_NOENT))
+                {
+                  g_free (path);
+                  path = NULL;
+                }
+
+              /* cleanup (error is reused for file monitoring below) */
+              g_clear_error (&error);
+              g_free (dir);
+            }
+
+          /* we managed to get the final target */
+          if (path != NULL)
+            {
+              file->monitor_location = g_file_new_for_path (path);
+              g_free (path);
+            }
+          /* fall back to the original location */
+          else
+            file->monitor_location = g_object_ref (file->location);
         }
       else
-        location = g_object_ref (file->location);
+        file->monitor_location = g_object_ref (file->location);
 
-      file->monitor = g_file_monitor_file (location,
+      file->monitor = g_file_monitor_file (file->monitor_location,
 #if GLIB_CHECK_VERSION (2, 56, 2)
                                            G_FILE_MONITOR_WATCH_HARD_LINKS |
 #endif
                                            G_FILE_MONITOR_WATCH_MOVES,
                                            NULL, &error);
-      g_object_unref (location);
 
       /* inform the user */
       if (error != NULL)
         {
-          g_message ("File monitoring is disabled for file '%s': %s",
-                     mousepad_file_get_path (file), error->message);
+          path = mousepad_util_get_display_path (file->location);
+          g_message ("File monitoring is disabled for file '%s': %s", path, error->message);
+          g_free (path);
           g_error_free (error);
         }
       /* watch file for changes */
@@ -282,10 +391,15 @@ mousepad_file_set_monitor (gpointer data)
 void
 mousepad_file_set_location (MousepadFile *file,
                             GFile        *location,
-                            gboolean      real)
+                            gint          type)
 {
+  GFileInfo *fileinfo;
+
   g_return_if_fail (MOUSEPAD_IS_FILE (file));
 
+  /* update location state */
+  file->temporary = (type == MOUSEPAD_LOCATION_VIRTUAL);
+
   /* set location */
   if (file->location == NULL && location != NULL)
     {
@@ -293,7 +407,7 @@ mousepad_file_set_location (MousepadFile *file,
 
       /* mark the document as modified (and therefore savable) in case of a new, empty
        * but localized document */
-      if (! g_file_query_exists (location, NULL))
+      if (! mousepad_util_query_exists (location, TRUE))
         gtk_text_buffer_set_modified (file->buffer, TRUE);
     }
   /* reset location */
@@ -311,10 +425,21 @@ mousepad_file_set_location (MousepadFile *file,
     }
 
   /* not a virtual change, such as when trying to save as */
-  if (real)
+  if (type == MOUSEPAD_LOCATION_REAL)
     {
-      /* this is a definitive location */
-      file->temporary = FALSE;
+      /* update read-only status */
+      if (mousepad_util_query_exists (location, TRUE)
+          && G_LIKELY (fileinfo = g_file_query_info (location, G_FILE_ATTRIBUTE_ACCESS_CAN_WRITE,
+                                                     G_FILE_QUERY_INFO_NONE, NULL, NULL)))
+        {
+          mousepad_file_set_read_only (file,
+            ! g_file_info_get_attribute_boolean (fileinfo, G_FILE_ATTRIBUTE_ACCESS_CAN_WRITE));
+          g_object_unref (fileinfo);
+        }
+      else if (mousepad_util_get_path (location) == NULL)
+        mousepad_file_set_read_only (file, TRUE);
+      else
+        mousepad_file_set_read_only (file, FALSE);
 
       /* activate file monitoring with a delay, to not consider our own saving as
        * external modification after a save as */
@@ -322,11 +447,8 @@ mousepad_file_set_location (MousepadFile *file,
                      mousepad_file_set_monitor, mousepad_util_source_autoremove (file));
 
       /* send a signal that the name has been changed */
-      g_signal_emit (file, file_signals[LOCATION_CHANGED], 0, file->location);
+      g_signal_emit (file, file_signals[LOCATION_CHANGED], 0, location);
     }
-  /* toggle location state */
-  else
-    file->temporary = ! file->temporary;
 }
 
 
@@ -599,7 +721,16 @@ mousepad_file_open (MousepadFile  *file,
   if ((autosave_uri = mousepad_object_get_data (file->location, "autosave-uri")) != NULL)
     location = g_file_new_for_uri (autosave_uri);
   else
-    location = g_object_ref (file->location);
+    {
+      location = g_object_ref (file->location);
+
+      /* update monitor location in case of a symlink (really useful only on reload,
+       * but not very costly) */
+      if (file->monitor != NULL && (
+            file->symlink || (file->symlink = mousepad_util_is_symlink (file->location))
+          ))
+        mousepad_file_set_monitor (file);
+    }
 
   /* if the file does not exist and this is allowed, no problem */
   if (! g_file_load_contents (location, NULL, &contents, &file_size, &etag, error)
@@ -827,15 +958,33 @@ mousepad_file_replace_contents (MousepadFile      *m_file,
   if (m_file->monitor != NULL)
     g_signal_handlers_block_by_func (m_file->monitor, mousepad_file_monitor_changed, m_file);
 
+  /*
+   * Prevent g_file_replace_contents() from returning G_IO_ERROR_WRONG_ETAG when the
+   * link target has been removed.
+   * See https://gitlab.gnome.org/GNOME/glib/-/issues/2466
+   */
+  if (mousepad_util_is_symlink (m_file->location)
+      && ! mousepad_util_query_exists (m_file->location, TRUE))
+    etag = NULL;
+
   /* replace contents */
   succeed = g_file_replace_contents (file, contents, length, etag, make_backup,
                                      flags, new_etag, cancellable, error);
 
-  /* reactivate file monitoring with a delay, to not consider our own saving as
-   * external modification */
   if (m_file->monitor != NULL)
-    g_timeout_add (MOUSEPAD_SETTING_GET_UINT (MONITOR_DISABLING_TIMER),
-                   mousepad_file_monitor_unblock, mousepad_util_source_autoremove (m_file));
+    {
+      /* update monitor location in case of a symlink */
+      if (succeed && (
+            m_file->symlink || (m_file->symlink = mousepad_util_is_symlink (m_file->location))
+         ))
+        g_timeout_add (MOUSEPAD_SETTING_GET_UINT (MONITOR_DISABLING_TIMER),
+                       mousepad_file_set_monitor, mousepad_util_source_autoremove (m_file));
+      /* reactivate file monitoring with a delay, to not consider our own saving as
+       * external modification */
+      else
+        g_timeout_add (MOUSEPAD_SETTING_GET_UINT (MONITOR_DISABLING_TIMER),
+                       mousepad_file_monitor_unblock, mousepad_util_source_autoremove (m_file));
+    }
 
   return succeed;
 }
@@ -920,7 +1069,7 @@ mousepad_file_prepare_save_contents (MousepadFile  *file,
           case MOUSEPAD_EOL_UNIX:
             if (contents[length - 1] != '\n')
               {
-                contents = g_realloc (contents, length + 2);
+                contents = g_renew (gchar, contents, length + 2);
                 contents[length] = '\n';
                 length++;
                 eol = "\n";
@@ -930,7 +1079,7 @@ mousepad_file_prepare_save_contents (MousepadFile  *file,
           case MOUSEPAD_EOL_MAC:
             if (contents[length - 1] != '\r')
               {
-                contents = g_realloc (contents, length + 2);
+                contents = g_renew (gchar, contents, length + 2);
                 contents[length] = '\r';
                 length++;
                 eol = "\r";
@@ -940,7 +1089,7 @@ mousepad_file_prepare_save_contents (MousepadFile  *file,
           case MOUSEPAD_EOL_DOS:
             if (contents[length - 1] != '\n' || (length > 1 && contents[length - 2] != '\r'))
               {
-                contents = g_realloc (contents, length + 3);
+                contents = g_renew (gchar, contents, length + 3);
                 contents[length] = '\r';
                 contents[length + 1] = '\n';
                 length += 2;
@@ -984,42 +1133,38 @@ mousepad_file_save (MousepadFile  *file,
     return FALSE;
 
   /* write the buffer to the file */
-  if (mousepad_file_replace_contents (file, file->location, contents, length,
-                                      (file->temporary || forced) ? NULL : file->etag,
-                                      MOUSEPAD_SETTING_GET_BOOLEAN (MAKE_BACKUP),
-                                      G_FILE_CREATE_NONE, &etag, NULL, error))
-    {
-      /* update etag */
-      g_free (file->etag);
-      file->etag = etag;
-
-      /* update read-only status in case of save as */
-      if (G_UNLIKELY (file->temporary))
-        mousepad_file_set_read_only (file, FALSE);
-
-      /* add last eol if needed */
-      if (eol != NULL)
-        {
-          gtk_text_buffer_get_end_iter (file->buffer, &iter);
-          gtk_text_buffer_insert (file->buffer, &iter, eol, -1);
-          g_free (eol);
-        }
-
-      /* cleanup */
-      g_free (contents);
-    }
-  else
+  if (! mousepad_file_replace_contents (file, file->location, contents, length,
+                                        (file->temporary || forced) ? NULL : file->etag,
+                                        MOUSEPAD_SETTING_GET_BOOLEAN (MAKE_BACKUP),
+                                        G_FILE_CREATE_NONE, &etag, NULL, error))
     {
       g_free (contents);
+      g_free (eol);
+
       return FALSE;
     }
 
+  /* update etag */
+  g_free (file->etag);
+  file->etag = etag;
+
+  /* add last eol if needed */
+  if (eol != NULL)
+    {
+      gtk_text_buffer_get_end_iter (file->buffer, &iter);
+      gtk_text_buffer_insert (file->buffer, &iter, eol, -1);
+      g_free (eol);
+    }
+
   /* everything has been saved */
   gtk_text_buffer_set_modified (file->buffer, FALSE);
 
   /* re-guess the filetype which could have changed */
   mousepad_file_set_language (file, NULL);
 
+  /* cleanup */
+  g_free (contents);
+
   return TRUE;
 }
 
@@ -1228,7 +1373,7 @@ mousepad_file_autosave_save_sync (MousepadFile *file)
   GtkWindow  *window;
   GError    **perror = NULL;
   GError     *error = NULL;
-  gchar      *contents;
+  gchar      *contents = NULL;
   gsize       length;
 
   /* file already saved */
@@ -1254,8 +1399,9 @@ mousepad_file_autosave_save_sync (MousepadFile *file)
     }
 
   /* save contents */
-  if (! g_file_replace_contents (file->autosave_location, contents, length, NULL, FALSE,
-                                 G_FILE_CREATE_NONE, NULL, NULL, perror)
+  if (contents != NULL
+      && ! g_file_replace_contents (file->autosave_location, contents, length, NULL, FALSE,
+                                    G_FILE_CREATE_NONE, NULL, NULL, perror)
       && perror != NULL)
     {
       window = gtk_application_get_active_window (GTK_APPLICATION (g_application_get_default ()));
